(* $Id: compile.ml,v 1.13 1998/02/16 07:07:02 garrigue Exp $ *)

#open"misc";;
#open"common";;
#open"define";;
#open"untyped";;

type var_alloc =
    { mutable size: int ;
      mutable table: (ident, int) map__t ;
      mutable defined: ident set__t ;
      mutable needed: ident set__t }
;;

let empty_set = set__empty compare_id ;;
let empty_table = map__empty compare_id ;;

let empty_alloc () =
  { size = 0 ; table = empty_table ;
    defined = empty_set ; needed = empty_set } ;;
let new_alloc alloc =
  { size = 0 ; table = empty_table ;
    defined = alloc.defined ; needed = empty_set } ;;

let add_alloc alloc id =
  let index = alloc.size in
  alloc.defined <- set__add id alloc.defined;
  alloc.table <- map__add id index alloc.table;
  alloc.size <- index + 1;
  index
;;

let find_alloc alloc id =
  if set__mem id alloc.defined then
    try map__find id alloc.table
    with Not_found ->
      alloc.needed <- set__add id alloc.needed;
      add_alloc alloc id
  else
    raise Not_found
;;
  
let repr_constant = function
    Cint n -> obj__repr n
  | Cchar c -> obj__repr c
  | Cfloat x -> obj__repr x
;;

type pattern =
    CPid of int
  | CPint of int
  | CPconst of obj__obj
  | CPblock of int * pattern vect
  | CPany
;;

let rec pattern alloc = function
    UPid id ->
      let n = add_alloc alloc id in
      CPid n
  | UPconst (Cfloat x) ->
      CPconst (obj__repr x)
  | UPconst c ->
      CPint (obj__magic_obj (repr_constant c))
  | UPblock (tag, []) ->
      CPint (obj__magic_obj (obj__obj_block tag 0))
  | UPblock (tag, l) ->
      let cv = map_vect (pattern alloc) (vect_of_list l) in
      begin try
	let obj = obj__obj_block tag (vect_length cv) in
	for i = 0 to vect_length cv - 1 do
	  let field =
	    match cv.(i) with
		CPint x -> obj__repr x
	      | CPconst obj -> obj
	      | _ -> raise Not_found
	  in obj__set_obj_field obj i field
	done;
	CPconst obj
      with
	Not_found -> CPblock (tag, cv)
      end
  | UPany ->
      CPany
;;

(* from now on we need fast vector accesses *)
#open"fvect";;
#open"obj";;

let matching alloc exn pat =
  let cpat = pattern alloc pat in
  let rec match_rec = function
      CPany ->
	(fun _ _ -> ())
    | CPid n ->
	(fun env obj -> env.(n) <- obj)
    | CPint x ->
	(fun _ obj -> if x <> (magic_obj obj) then raise exn)
    | CPconst x ->
	(fun _ obj -> if x <> obj then raise exn)
    | CPblock (tag, arr) ->
	(fun env obj ->
	  let len = vect_length arr in
	  if obj_size obj <> len or obj_tag obj <> tag then raise exn;
	  for i = 0 to len - 1 do
	    match_rec arr.(i) env (magic_obj obj).(i)
	  done)
  in
  match_rec cpat
;;

exception Match_failure;;

let return obj env = obj ;;

let rec command alloc = function
    UEexpr e ->
      let code = expression alloc e in
      (fun env -> code env; ())
  | UEval (UPid id, e, err) ->
      let code = expression alloc e in
      let i = add_alloc alloc id in
      (fun env -> env.(i) <- code env)
  | UEval (pat, e, err) ->
      let mtc = matching alloc (Match_error err) pat
      and code = expression alloc e in
      (fun env -> mtc env (code env))
  | UEfun l ->
      let offset = alloc.size in
      let v = vect_of_list l in
      (* Allocate locally only if not defined globally *)
      begin try
	let orefs = map_vect (fun (id,e) -> map__find id !global_env) v in
	let codes = map_vect (fun (id, e) -> expression alloc e) v
	and last = vect_length v -  1 in
	(fun env ->
	  for i = 0 to last do
	    orefs.(i) := codes.(i) env
	  done)
      with
	Not_found ->
	  do_vect (fun (id, e) -> let _ = add_alloc alloc id in ()) v;
	  let codes = map_vect (fun (id, e) -> expression alloc e) v
	  and last = vect_length v -  1 in
	  (fun env ->
	    for i = 0 to last do
	      env.(i+offset) <- codes.(i) env
	    done)
      end
  | UEvar (id, e) ->
      let code = expression alloc e in
      let i = add_alloc alloc id in
      (fun env -> env.(i) <- code env)

and expression alloc = function
    UEid id ->
      begin try
	let i = find_alloc alloc id in
	(fun env -> env.(i))
      with
	Not_found ->
	  try
	    let oref = map__find id !global_env in
	    (fun env -> !oref)
	  with
	    Not_found ->
	      failwith (id.name ^ " not found")
      end
  | UEconst c -> return (repr_constant c)
  | UEblock (tag, []) -> return (obj_block tag 0)
  | UEblock (tag, l) ->
      let codes = map_vect (expression alloc) (vect_of_list l) in
      let size = vect_length codes in
      (fun env ->
	let obj = obj_block tag size in
	for i = 0 to pred(size) do
	  set_obj_field obj i (codes.(i) env)
	done;
	obj)
  | UEapply (UEid{name="&"|"or" as op}, [e1;e2]) ->
      let code1 = expression alloc e1
      and code2 = expression alloc e2 in
      if op = "&" then
	(fun env -> repr(magic_obj(code1 env) & magic_obj(code2 env)))
      else
	(fun env -> repr(magic_obj(code1 env) or magic_obj(code2 env)))
  | UEapply (e, [e1]) ->
      let code = expression alloc e
      and code1 = expression alloc e1 in
      (fun env -> (magic_obj (code env)) (code1 env))
  | UEapply (e, [e1; e2]) ->
      let code = expression alloc e
      and code1 = expression alloc e1
      and code2 = expression alloc e2 in
      (fun env -> (magic_obj (code env)) (code1 env) (code2 env))
  | UEapply (e, l) ->
      let codes = map (expression alloc) l
      and code = expression alloc e in
      let rec apply env obj = function
	  [] -> obj
	| code :: l -> apply env ((magic_obj obj) (code env)) l
      in
      (fun env -> apply env (code env) codes)
  | UEfunct (ids, e) ->
      let alloc' = new_alloc alloc in
      do_list (fun id -> let _ = add_alloc alloc' id in ()) ids;
      let code = expression alloc' e in
      let needed = vect_of_list (set__elements alloc'.needed) in
      let source =  map_vect (find_alloc alloc) needed
      and dest = map_vect (find_alloc alloc') needed
      and last_copy = vect_length needed - 1
      and last_id = list_length ids - 1
      and size = alloc'.size in
      if last_id = 0 then
	if last_copy = -1 then
	  (magic
	   (fun env x ->
	     let new_env = (magic_obj (obj_block 0 size) : obj vect) in
	     new_env.(0) <- x;
	     code new_env)
	   : obj vect -> obj)
	else
	  (magic
	   (fun env x ->
	     let new_env = (magic_obj (obj_block 0 size) : obj vect) in
	     new_env.(0) <- x;
	     for i = 0 to last_copy do
	       new_env.(dest.(i)) <- env.(source.(i))
	     done;
	     code new_env)
	   : obj vect -> obj)
      else
      let pred_id = pred (last_id) in
      let rec copy_arg n l env =
	if n = 0 then
	  env.(0) <- l
	else begin
	  env.(n) <- obj_field l 0;
	  copy_arg (pred n) (obj_field l 1) env
	end in
      let code =
	if last_copy = -1 then
	  (magic
	   (fun env l x ->
	     let new_env = (magic_obj (obj_block 0 size) : obj vect) in
	     new_env.(last_id) <- x; copy_arg pred_id l new_env;
	     code new_env)
	   : obj vect -> obj -> obj)
	else
	  (magic
	   (fun env l x ->
	     let new_env = (magic_obj (obj_block 0 size) : obj vect) in
	     new_env.(last_id) <- x; copy_arg pred_id l new_env;
	     for i = 0 to last_copy do
	       new_env.(dest.(i)) <- env.(source.(i))
	     done;
	     code new_env)
	   : obj vect -> obj -> obj)
      in
      let rec make_code n code =
	if n = 0 then
	  (magic code : obj vect -> obj)
	else
	  make_code (pred n)
	    (magic (fun env l x -> code env (repr (x::l)))
	     : obj vect -> obj -> obj)
      in
      make_code pred_id code
  | UEseq ([], e) ->
      expression alloc e
  | UEseq (cmds1, UEseq (cmds2, e)) ->
      expression alloc (UEseq (cmds1 @ cmds2, e))
  | UEseq (cmds, e) ->
      let ccmds = map_vect (command alloc) (vect_of_list cmds) in
      let code = expression alloc e in
      let last = vect_length ccmds - 1 in
      if last = 0 then
	let ccmd = ccmds.(0) in (fun env -> ccmd env; code env)
      else
	(fun env ->
	  for i = 0 to last do ccmds.(i) env done;
	  code env)
  | UEcase (e, cases, err) ->
      let code = expression alloc e in
      let i, need_eval =
	try match e with
	  UEid id -> find_alloc alloc id, false
	| _ -> raise Not_found
	with Not_found -> add_alloc alloc (new_id "case"), true
      in
      let alloc' =
	{ defined = alloc.defined ; needed = alloc.needed ;
	  size = alloc.size ; table = alloc.table } in
      do_list
        (fun (pat,e) ->
	  let _ = matching alloc' Match_failure pat
	  and _ = expression alloc' e in ())
        cases;
      let new_needed = set__diff alloc'.needed alloc.needed in
      set__iter (fun id -> let _ = find_alloc alloc id in ()) new_needed;
      let needed = alloc.needed
      and size = alloc.size in
      let max_size = ref size in
      let make_code exn (pat, e) =
	alloc.size <- size;
	let mtc = matching alloc exn pat
	and code = expression alloc e in
	max_size := max alloc.size !max_size;
	(fun env -> mtc env env.(i); code env)
      in
      let cases, last_case = split_last cases in
      let codes = map (make_code Match_failure) cases
      and last_code = make_code (Match_error err) last_case in
      alloc.size <- !max_size;
      if not set__equal needed alloc.needed then failwith "compile__UEcase";
      let rec try_matching env = function
	  [] -> last_code env
	| coden :: l ->
	    try coden env with Match_failure -> try_matching env l
      in if need_eval then
	(fun env -> env.(i) <- code env; try_matching env codes)
      else
	(fun env -> try_matching env codes)
  | UEifthenelse (e, e1, e2) ->
      let code = expression alloc e
      and code1 = expression alloc e1
      and code2 = expression alloc e2 in
      (fun env ->
	if magic_obj (code env) then code1 env else code2 env)
  | UEset (id, e) ->
      let code = expression alloc e in
      begin try
	let i = find_alloc alloc id in
	(fun env -> repr (env.(i) <- code env))
      with Not_found ->
	try
	  let oref = map__find id !global_env in
	  (fun env -> repr(oref := code env))
	with
	  Not_found ->
	    failwith (id.name ^ " not found")
      end
  | UEfor (id, e1, dir, e2, e) ->
      let i = add_alloc alloc id in
      let code1 = expression alloc e1
      and code2 = expression alloc e2
      and code = expression alloc e in
      if dir = Upto then
      	(fun env -> repr(
	  for j = magic_obj (code1 env) to magic_obj (code2 env)
	  do
	    env.(i) <- repr j;
	    let _ = code env in ()
	  done))
      else
      	(fun env -> repr(
	  for j = magic_obj (code1 env) downto magic_obj (code2 env)
	  do
	    env.(i) <- repr j;
	    let _ = code env in ()
	  done))
  | UEwhile (e1, e2) ->
      let code1 = expression alloc e1
      and code2 = expression alloc e2 in
      (fun env ->
	repr(while magic_obj (code1 env) do let _ = code2 env in () done))
;;

let compile_expression e =
  let alloc = empty_alloc () in
  let code = expression alloc e in
  (fun () ->
    let env = make_vect alloc.size (repr ()) in
    code env)
;;

let compile_commands ucmds idents =
  let alloc = empty_alloc () in
  let codes = map (command alloc) ucmds in
  let env = make_vect alloc.size (repr ()) in
  do_list (fun code -> code env) codes;
  map 
    (fun id ->
      try vect_item env (map__find id alloc.table)
      with Not_found ->
	try !(map__find id !global_env)
	with Not_found -> failwith "compile__compile_commands")
    idents
;;

let prealloc_idents cmds =
  flat_map
    (function UEfun l -> map fst l | _ -> [])
    cmds
;;
